/**********************************************************************\
*        LongCoils.cpp computes pairs of long coils and magnets        *
*         with maximally flat induction: d2B/dz2 = 0 at z=0            *
*             by Marc Schaefer, aka Enthalpy, 12 Feb 2023              *
\**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

const double pi = 3.14159265359, twopi = 6.28318530718;
const double mu = 4e-7*pi; // >>>>> NEGLECT RELATIVE MU!! <<<<<
const double frac13 = 1.0/3.0, frac23 = 2.0/3.0;
const int stop = 40; // Iterations in dichotomy
const double eps = 1e-5; // Accepted rel diff

double H = 1.3/mu; // 1.3T typical for Nd magnets
double a = 0.0, b = 0.0, e = 0.0, R = 1.0; // a b e relative to R
double B0, d2B, d4B; // B0, d2B/dx2, d4B/dx4 at x=0. b != B != B0
double val; // The value typed by the user
double DropB[] = {1.0, 0.999, 0.998, 0.995, 0.99, 0.98, 0.95};
const int DropSz = sizeof (DropB)/sizeof (DropB[0]);
double DropZ[DropSz] = {0.0};


//******************************* display ******************************

void show (void) {
   printf (
      "  a (m)     b (m)     e (m)    R/2 (m)    R (m)     2R (m)\n"
      "%9.3e %9.3e %9.3e %9.3e %9.3e %9.3e\n",
       a*R, b*R, e*R, 0.5*R, R, 2*R // a b e relative to R
   );
   printf (
      " H (A/m)   B (T @0.0)  d2B/dx2 (T @R/2)  d4B/dx4 (T @R/2)\n"
      "%9.3e  %9.3e      %9.3e        %9.3e\n\n",
       H, B0, d2B*R*R/8.0, d4B*R*R*R*R/384.0
); }

double B (double z) ;

void Btenth (void) { // Variation of B at 0.1 0.2 0.3 0.4 0.5*R
   printf (
      "  B @0.0   @0.1*R   @0.2*R   @0.3*R   @0.4*R   @0.5*R\n"
      "%9.3e  %6.4f   %6.4f   %6.4f   %6.4f   %6.4f\n\n",
       B0, B (0.1)/B0, B (0.2)/B0, B (0.3)/B0, B (0.4)/B0, B (0.5)/B0
); }

void DropDisp (void) {
   printf ("relative B:    "); for (int i=1; i<DropSz; i++)
      printf ("%6.4f%s", DropB[i], (i>=DropSz-1 ? "\n" : "  ") );
   printf ("at positions:  "); for (int i=1; i<DropSz; i++)
      printf ("%6.4g%s", DropZ[i]*R, (i>=DropSz-1 ? "\n\n" : "  ") );
}

void help (void) {
   printf (
      "H <value in A/m>  -->  current per length unit\n"
      "a <value in m>    -->  half-distance between inner faces\n"
      "b <value in m>    -->  half-distance between outer faces\n"
      "e <value in m>    -->  coil or magnet length\n"
      "R <value in m>    -->  coil or magnet radius (not diameter)\n"
      "                       a, b, e scale then like R\n"
      "t                 -->  B at positions tenth or R\n"
      "d                 -->  positions where B drops\n"
      "?                 -->  display this help message\n"
      "q                 -->  exit this program\n\n"
); }


//*********************** compute magnetic field ***********************

inline double sqr (register double y) {return y*y;}

inline double dabs (register double y) {return (y > 0.0 ? y : -y);}
// #pragma intrinsic fabs needed a too big effort

inline double g (register double x, register double z)
   { return pow (1.0+sqr (x-z), -1.5) + pow (1.0+sqr (x+z), -1.5); }

double sum (double lo, double glo, double hi, double ghi, double z) {
   // Not Romberg. Unequal steps, tree sum, ends with Simpson
   double mi = 0.5*(lo+hi), gmi = g (mi, z), gme = 0.5*(glo+ghi);
return (
   dabs (gme-gmi) < eps*gmi ?
   (hi-lo)*(frac13*gme+frac23*gmi) : // Simpson
   sum (lo, glo, mi, gmi, z) + sum (mi, gmi, hi, ghi, z) // tree
); }

double B (double z) // B at offset z from center relative to R
   { return 0.5*mu*H*sum (a, g (a, z), b, g (b, z), z); }

void DropComp (void) { // Fills DropZ[] with z positions of predefined B
   int i, j; double lo, hi, mi;
   for (i=1; i<DropSz; i++) {
      lo = 0.0; hi = b*2.0; // Relative to R
         for (j = 0; j < stop; j++) // Dichotomy finds DropZ[i]
            if (mi = 0.5*(lo+hi), B (mi) > B0*DropB[i]) lo = mi;
            else hi = mi; // B must decrease with z, for right DropZ[]
         DropZ[i] = mi;
}  }

void field (void) { // Computes B0 d2B d4B from H a b e R
   double asqr = a*a, bsqr = b*b, Rsqr = R*R;

   B0   = B (0.0);
   d2B = 3.0*mu*H/Rsqr*(a*pow(1+asqr, -2.5)-b*pow(1+bsqr, -2.5));
   d4B = 120.0*mu*H/(Rsqr*Rsqr)*( // 120 is unsure!
            b*(0.75-bsqr)*pow(1.0+bsqr, -4.5)
          - a*(0.75-asqr)*pow(1.0+asqr, -4.5)
);       }

void UpdH (void) {
   if (val <= 0.0) {
      printf ("Need 0 < H. Nothing done.\n\n"); 
return;}
   H = val;
   if (a <= 0.0) return;
   field (); show ();
}


//************************* compute dimensions *************************

double f (double x) { // f (a) == f (b) for d2B/dx2 == 0.0
   // f (x) = x^2/(1+x^2)^5, see htm 30 Jan 2023
   double sqrx = x*x, sqrx1 = sqrx+1.0, sqrsqrx1 = sqrx1*sqrx1;
return (sqrx/(sqrx1*sqrsqrx1*sqrsqrx1));}

void UpdA (void) {
   if (val < 0.0 || R <= 2.0*val)
      {printf ("Need 0 <= a < R/2. Nothing done.\n\n"); return;}
   a = val/R; double fa = f (a); // a b e relative to R
   // Bounds for b. Use x^2/(1+x^2)^5 < x^-8
   double lo = 0.5, hi = pow (fa, -1.0/8.0), mi;
   for (int i = 0; i < stop; i++) // Dichotomy finds b from a
      if (mi = 0.5*(lo+hi), f (mi) > fa) lo = mi;
      else hi = mi;
   b = mi; e = b-a;
   field (); show ();
}

void UpdB (void) {
   if (2.0*val <= R)
      {printf ("Need R/2 < b. Nothing done.\n\n"); return;}
   b = val/R; double fb = f (b); // a b e relative to R
   double lo = 0.0, hi = 0.5, mi;
   for (int i = 0; i < stop; i++) // Dichotomy finds a from b
      if (mi = 0.5*(lo+hi), f (mi) > fb) hi = mi;
      else lo = mi;
   a = mi; e = b-a;
   field (); show ();
}

void UpdE (void) {
   if (val <= 0.0)
      {printf ("Need 0 < e. Nothing done.\n\n"); return;}
   e = val/R; // a b e relative to R
   double lo = 0.0, hi = 0.5, mi;
   for (int i = 0; i < stop; i++) // Dichotomy finds a b from e
      if (mi = 0.5*(lo+hi), f (mi) > f (mi+e)) hi = mi;
      else lo = mi;
   a = mi; b = mi+e;
   field (); show ();
}

void UpdR (void) {
   if (val <= 0.0)
      {printf ("Need 0 < R. Nothing done.\n\n"); return;}
   R = val; // a b e scale with R, keep their relative values
   if (a <= 0.0) return;
   field (); show ();
}


//************************** command interpret *************************

int main () {
   char command[81], *comm; int i;

   printf (
      "\nWelcome! This program by Marc Schaefer, aka Enthalpy,\n"
      "computes dimensions for pairs of long coils or magnets\n"
      "that flatten the induction at the centre: d2B/dx2 = 0\n\n"

      "This is no professional quality. Use at your own risks.\n\n"

      "Units are metric: H in A/m - B in T - a, b, e, R in m\n"
      "? displays the help, q quits.\n\n"
   );

   for (; ; ) {
      printf ("command: "); (void) gets (command);
      for (i = 0; command[i]; i++) command[i] = tolower (command[i]);
      for (comm = command; *comm && isspace (*comm); comm++) ;
      switch (*comm) {
         case 'h': if (sscanf (comm + 1, "%lf", &val ) != 1) help ();
            else UpdH (); break;
         case 'a': if (sscanf (comm + 1, "%lf", &val ) != 1) help ();
            else UpdA (); break;
         case 'b': if (sscanf (comm + 1, "%lf", &val ) != 1) help ();
            else UpdB (); break;
         case 'e': if (sscanf (comm + 1, "%lf", &val ) != 1) help ();
            else UpdE (); break;
         case 'r': if (sscanf (comm + 1, "%lf", &val ) != 1) help ();
            else UpdR (); break;
         case 't': Btenth (); break;
         case 'd': if (a <= 0.0) break; DropComp (); DropDisp (); break;
         case 'q': printf
            ("Bye-bye... Type exit to close the console\n");
exit (0);
         case '?': default: help (); break;
}   }   }
